% parametric squared euclidean distance
% with squared diagonal metric 
classdef ParametricDistance < handle
    properties (GetAccess='public',SetAccess = 'private')
        bInitialized = false;
        Lambda       = [];
        D            = 0;
    end
    
    methods
        % constructor
        function obj = ParametricDistance(NrDimensions)
            obj        = obj@handle();            
            obj.Lambda = ones(NrDimensions,1);
            obj.D      = NrDimensions;
            obj.bInitialized = true;
        end
        
        function Ret = GetLambda(me)
            Ret = me.Lambda;
            return
        end

        function Ret = GetParameter(me)
            Ret = me.Lambda(:); 
        end
       
        
        function [Ok] = SetParameter(me,LambdaVector)
            assert(numel(LambdaVector)==me.D)
            vLam = LambdaVector(:);
            if(1) % to check gradient
                vLam(vLam<0) = 1e-4./me.D; % this should be not to small to avoid an invalid discriminative metric in dgtm
                me.Lambda    = vLam./sum(vLam);
            else
                me.Lambda = vLam;
            end
            Ok = 1;
        end
        
        function Value = subsref(me,Indices)
            Value = [];
            NumInd = numel(Indices);
            if(NumInd==1) % direct access 
                Subs       = Indices.subs;
                NumIndSubs = numel(Subs);
                switch Indices.type
                    case '()'     % indexing                        
                        if(NumIndSubs==2)
                            Value = me.CalculateDistance(Subs{1},Subs{2});
                        else
                            ME = MException('Class:ParametricDistance:Error','number of parameters must be two');
                            throw(ME);
                        end
                    case '.'     % class member
                        Value = me.(Subs);
                end
            else % should be a function call
                if(strncmp(Indices(1).type,'.',1))   
                    Value = builtin('subsref', me, Indices); 
                end
            end
        end       
        
        % input D x N, D x M with D number of dimensions equal length of lambda
        function Dist = CalculateDistance(me,X,Y)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end            
            N             = size(X,1);
            iNrPrototypes = size(Y,1);
            weucHelp = @(Diff)reshape(sum(bsxfun(@times,Diff,me.Lambda).^2),N,iNrPrototypes); % squared euclidean distance with squared (!) weights            
            DistHelp = @(X,Y)weucHelp(reshape(permute(permute(repmat(Y, [1,1,N]),[3 2 1]) - repmat(X,[1,1,iNrPrototypes]),[2 1 3]),me.D,N*iNrPrototypes) ); 
            Dist = DistHelp(X,Y);            
        end
        
        function Dist = CalculateDistanceDerivativeLambda(me,X,Y,Degree)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end        
            
            switch Degree
                case 1
                    Dist = 2* bsxfun(@minus,X,Y)'.^2 .* me.Lambda;
                case 2
                    Dist = 2* bsxfun(@minus,X,Y)'.^2;
                otherwise
                    throw(MException('Class:ParametricDistance:Error','only first and second derivative supported'));                    
            end
        end        

        function Dist = CalculateDistanceDerivativeLambdaMulti(me,X,Y,Degree)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end        
            
            switch Degree
                case 1
                    Dist = 2* bsxfun(@times,bsxfun(@minus,X,Y)'.^2, me.Lambda); 
                case 2
                    Dist = 2* bsxfun(@minus,X,Y)'.^2;
                otherwise
                    throw(MException('Class:ParametricDistance:Error','only first and second derivative supported'));                    
            end
        end        
        
        
        function Dist = CalculateDistanceDerivative2ndArg(me,X,Y,Degree)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end        
            
            switch Degree
                case 1
                    Dist = -2* bsxfun(@minus,X,Y)' .* me.Lambda.^2;
                case 2
                    Dist = +2* me.Lambda.^2;
                otherwise
                    throw(MException('Class:ParametricDistance:Error','only first and second derivative supported'));                    
            end
        end                
        
        function Dist = CalculateDistanceDerivative2ndArgMultiY(me,X,Y,Degree)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end        
            
            switch Degree
                case 1
                    Dist = -2 * bsxfun(@times,bsxfun(@minus,X,Y)', me.Lambda.^2);
                case 2
                    Dist = repmat(+2* me.Lambda.^2,1,size(Y,1));
                otherwise
                    throw(MException('Class:ParametricDistance:Error','only first and second derivative supported'));                    
            end
        end                

        function Dist = CalculateDistanceDerivative2ndArgMultiX(me,X,Y,Degree)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end        
            
            switch Degree
                case 1
                    Dist = -2 * bsxfun(@times,bsxfun(@minus,X,Y)', me.Lambda.^2);
                case 2
                    Dist = repmat(+2* me.Lambda.^2,1,size(X,1));
                otherwise
                    throw(MException('Class:ParametricDistance:Error','only first and second derivative supported'));                    
            end
        end                

        function Dist = CalculateDistanceDerivative2ndArgMultiXY(me,X,Y)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end        
            N             = size(X,1);
            iNrPrototypes = size(Y,1);
            X=bsxfun(@times,X,me.Lambda.^2');
            Y=bsxfun(@times,Y,me.Lambda.^2');
            Dist = 2 * (permute(repmat(Y,[1,1,N]),[3 2 1]) - repmat(X,[1,1,iNrPrototypes]));
        end                
        
    end
end